/*
    THELP32 - Windows 95 interface for Tool Help
    Copyright (c) 1995-2003 Alessandro Scotti
    http://www.walkofmind.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "thelp32.h"

#define THERR_SUCCESS       0
#define THERR_BADVERSION    -1
#define THERR_FAILURE       -2

// Type definitions for pointers to call tool help functions.
typedef BOOL (WINAPI *MODULEWALK)(HANDLE hSnapshot, LPMODULEENTRY32 lpme);
typedef BOOL (WINAPI *THREADWALK)(HANDLE hSnapshot, LPTHREADENTRY32 lpte);
typedef BOOL (WINAPI *PROCESSWALK)(HANDLE hSnapshot, LPPROCESSENTRY32 lppe);
typedef HANDLE (WINAPI *CREATESNAPSHOT)(DWORD dwFlags, DWORD th32ProcessID);

// Pointers to toolhelp functions (Windows 95 only)
static CREATESNAPSHOT   pCreateToolhelp32Snapshot   = NULL;
static MODULEWALK       pModule32First              = NULL;
static MODULEWALK       pModule32Next               = NULL;
static PROCESSWALK      pProcess32First             = NULL;
static PROCESSWALK      pProcess32Next              = NULL;
static THREADWALK       pThread32First              = NULL;
static THREADWALK       pThread32Next               = NULL;

/*----------------------------------------------------------

    Initializes the Tool Help library.

    Input:
        none
    Output:
        result

----------------------------------------------------------*/
int WINAPI thInit( void )
{
    BYTE    bVerHi;
    BYTE    bVerLo;
    HANDLE  hKernel;

    // Check version
    bVerLo = (GetVersion() & 0xFF00) >> 8;
    bVerHi = (GetVersion() & 0x00FF);
    if( (bVerHi != 03) || (bVerLo != 95) )
        return( THERR_BADVERSION );

    // Get the module handle of the kernel
    hKernel = GetModuleHandle( "KERNEL32.DLL" );
    if( hKernel == NULL )
        return( THERR_FAILURE );

    // Get the function addresses
    pCreateToolhelp32Snapshot = (CREATESNAPSHOT)GetProcAddress(hKernel,
        "CreateToolhelp32Snapshot");
    pModule32First  = (MODULEWALK)GetProcAddress(hKernel, "Module32First");
    pModule32Next   = (MODULEWALK)GetProcAddress(hKernel, "Module32Next");
    pProcess32First = (PROCESSWALK)GetProcAddress(hKernel, "Process32First");
    pProcess32Next  = (PROCESSWALK)GetProcAddress(hKernel, "Process32Next");
    pThread32First  = (THREADWALK)GetProcAddress(hKernel, "Thread32First");
    pThread32Next   = (THREADWALK)GetProcAddress(hKernel, "Thread32Next");

    return( THERR_SUCCESS );
}

/*----------------------------------------------------------

    Terminates the Tool Help library.

    Input:
        none
    Output:
        result

----------------------------------------------------------*/
int WINAPI thTerm( void )
{
    return( THERR_SUCCESS );
}

/*----------------------------------------------------------

    Tool Help functions.

----------------------------------------------------------*/
HANDLE WINAPI thCreateSnapshot( DWORD dwFlags, DWORD thProcessId )
{
    if( pCreateToolhelp32Snapshot == NULL )
        return( NULL );
    return( (*pCreateToolhelp32Snapshot)(dwFlags,thProcessId) );
}

BOOL WINAPI thProcessFirst( HANDLE hSnapshot, LPPROCESSENTRY32 lpPE )
{
    if( pProcess32First == NULL )
        return( FALSE );
    return( (*pProcess32First)(hSnapshot,lpPE) );
}

BOOL WINAPI thProcessNext( HANDLE hSnapshot, LPPROCESSENTRY32 lpPE )
{
    if( pProcess32Next == NULL )
        return( FALSE );
    return( (*pProcess32Next)(hSnapshot,lpPE) );
}

