/*
    WPX - DLL for hook functions and WordPad extensions
    Copyright (c) 1995-2003 Alessandro Scotti
    http://www.walkofmind.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#define STRICT
#define INC_OLE2
#include <windows.h>
#include <windowsx.h>
#include <richedit.h>
#include <richole.h>
#include <commctrl.h>
#include "wpx.h"
#include "wpxdll.rh"

/*----------------------------------------------------------

    Instance data (not shared).

----------------------------------------------------------*/
static HINSTANCE        hInst;                  // Handle of module
static HWND             hAppWnd;                // Handle of control window
static HHOOK            hShlHook = NULL;        // Shell hook handle
static HHOOK            hMsgHook = NULL;        // Message hook handle

static HMENU            hMenuTools = NULL;      // Handle of new menu

/* WordPad information */
static WNDPROC          pfOldWndProc = NULL;// Address of subclassed WndProc
static HMENU            hMenuMain = NULL;   // Handle of main menu
static HWND             hWndMain = NULL;    // Handle of main window
static HWND             hWndText = NULL;    // Handle of the RICHEDIT control
static HWND             hWndStatus = NULL;  // Handle of the status bar
static COLORREF         rgbTextBackground;  // Text background color
static COLORREF         rgbCustomColors[16];// Custom colors for ChooseColor

/*
    Classifies a character.
*/
DWORD subCharClassify( CHAR c )
{
    if( (c==' ') || (c=='\t') )
        return( CCLS_BLANK );
    if( (c=='\n') || (c=='\r') )
        return( CCLS_BREAK );
    if( IsCharAlphaNumeric(c) )
        return( CCLS_WORD );
    return( CCLS_OTHER );
}

/*
    Stream callback for the RichEdit control.

    Input:
        dwCookie    = (user defined) address of WPXTEXTINFO structure
        pbBuf       = pointer to text buffer
        cbBuf       = size of text buffer
        pcbOut      = pointer to LONG to receive number of chars actually read
    Output:
        number of character processed
*/
DWORD CALLBACK subStreamOutProc( DWORD dwCookie,
    LPBYTE pbBuf, LONG cbBuf, LONG *pcbOut )
{
    WPXTEXTINFO *   pWTI;
    WORD            wPosProgress;

    *pcbOut = cbBuf;
    pWTI = (WPXTEXTINFO *)dwCookie;

    while( cbBuf-- > 0 ) {
        /* Bump progress bar if needed */
        pWTI->dwCurChar++;
        if( pWTI->dwProgressStepSize > 0 )
            if( (pWTI->dwCurChar % pWTI->dwProgressStepSize) == 0 )
                SendMessage( pWTI->hWndProgress, PBM_STEPIT, 0, 0 );
        /* Get text info */
        switch( subCharClassify( (CHAR)*pbBuf++ ) ) {
            case CCLS_WORD:
                if( !pWTI->boInWord )
                    pWTI->dwWords++;
                if( !pWTI->boInPara )
                    pWTI->dwParagraphs++;
                pWTI->boInWord = TRUE;
                pWTI->boInPara = TRUE;
                break;
            case CCLS_BREAK:
                pWTI->boInPara = FALSE;
                /* Fall thru */
            case CCLS_BLANK:
                pWTI->dwLengthNoBlanks--;
                /* Fall thru */
            case CCLS_OTHER:
                pWTI->boInWord = FALSE;
                break;
        }
    }

    return( *pcbOut );
}

/*
    Enumeration callback for EnumChildWindows.
*/
BOOL CALLBACK subEnumChildProc( HWND hWnd, LPARAM lParam )
{
    CHAR    szBuf[128];
    RECT    stRC;
    DWORD * lpMaxX;

    if( GetClassName( hWnd, szBuf, sizeof(szBuf) ) > 0 ) {
        if( lstrcmp( szBuf, RICHEDITCLASSNAME ) == 0 ) {
            /* We have found the text control */
            hWndText = hWnd;
        }
        else if( lstrcmp( szBuf, AFXBARCLASSNAME ) == 0 ) {
            /* Maybe the status bar, check Y coordinate */
            lpMaxX = (DWORD *)lParam;
            GetWindowRect( hWnd, &stRC );
            if( stRC.bottom > *lpMaxX ) {
                /* Best candidate so far */
                hWndStatus = hWnd;
                *lpMaxX = stRC.bottom;
            }
        }
    }

    return( TRUE );
}

/*
    Displays a resource string inside the status bar.
*/
void subStatusText( DWORD dwIdx )
{
    CHAR    szBuf[80] = "";

    if( hWndStatus != NULL ) {
        if( dwIdx > 0 )
            if( LoadString( hInst, dwIdx, szBuf, sizeof(szBuf) ) == 0 )
                return;
        SendMessage( hWndStatus, WM_SETTEXT, 0, (LPARAM)szBuf );
    }
}

/*
    Converts a DWORD value into a string and updates
    the specified dialog item.
*/
void subTextInfo( HWND hWnd, DWORD dwItem, DWORD dwValue )
{
    CHAR    szBuf1[16];
    CHAR    szBuf2[16];
    DWORD   dwLen = 0;
    DWORD   dwIdx;

    do {
        szBuf1[dwLen++] = (dwValue % 10) + '0';
        dwValue /= 10;
    } while( dwValue > 0 );
    for( dwIdx=0; dwIdx<dwLen; dwIdx++ )
        szBuf2[dwIdx] = szBuf1[dwLen-dwIdx-1];
    szBuf2[dwIdx] = 0;

    SendDlgItemMessage( hWnd, dwItem, WM_SETTEXT, 0, (LPARAM)szBuf2 );
}

/*----------------------------------------------------------

    Text information dialog box.

----------------------------------------------------------*/
BOOL CALLBACK TextInfoDlgProc( HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam )
{
    HWND            hWndProgress;
    HWND            hWndOkButton;
    WPARAM          wStreamFlags;
    HCURSOR         hCursor;
    CHARRANGE       stCR;
    EDITSTREAM      stES;
    WPXTEXTINFO     stWTI;
    LPRICHEDITOLE   lpREO;

    NOREF(wParam);
    NOREF(lParam);

    switch( wMsg ) {
        /* Initialize dialog box */
        case WM_INITDIALOG:
            /* Setup dialog box */
            hWndProgress = GetDlgItem( hWnd, IDC_PROGRESS );
            hWndOkButton = GetDlgItem( hWnd, IDOK );
            ShowWindow( hWndOkButton, SW_HIDE );
            SendMessage( hWndProgress, PBM_SETPOS, 0, 0 );
            SendMessage( hWndProgress, PBM_SETSTEP, 1, 0 );
            SendMessage( hWndProgress, PBM_SETRANGE, 0,
                MAKELPARAM( 0, PROGRESS_STEPS ) );
            /* Post a message and return (so that dialog is displayed) */
            PostMessage( hWnd, PM_DLGGETTEXTINFO, 0, 0 );
            return( TRUE );
        /* Get text information */
        case PM_DLGGETTEXTINFO:
            /* First, get a couple of useful handles */
            hWndProgress = GetDlgItem( hWnd, IDC_PROGRESS );
            hWndOkButton = GetDlgItem( hWnd, IDOK );
            /* Set cursor to hourglass shape */
            hCursor = GetCursor();
            SetCursor( LoadCursor( NULL, MAKEINTRESOURCE(IDC_WAIT) ) );
            /* Get info about current selection */
            if( SendMessage(hWndText,EM_SELECTIONTYPE,0,0) != SEL_EMPTY ) {
                DWORD   dwLineMax, dwLineMin;

                wStreamFlags = SF_TEXT | SFF_SELECTION;
                SendMessage( hWndText, EM_EXGETSEL, 0, (LPARAM)&stCR );
                stWTI.dwLength = stCR.cpMax - stCR.cpMin;
                dwLineMax = SendMessage(
                    hWndText, EM_EXLINEFROMCHAR, 0, stCR.cpMax-1 );
                dwLineMin = SendMessage(
                    hWndText, EM_EXLINEFROMCHAR, 0, stCR.cpMin );
                stWTI.dwLines = dwLineMax - dwLineMin + 1;
            }
            else {
                /* No selection */
                wStreamFlags = SF_TEXT;
                stWTI.dwLength = SendMessage( hWndText, WM_GETTEXTLENGTH, 0, 0 );
                stWTI.dwLines = SendMessage( hWndText, EM_GETLINECOUNT, 0, 0 );
            }
            /* Get OLE information */
            stWTI.dwOLEObjects = 0;
            stWTI.dwOLELinks = 0;
            if( SendMessage( hWndText, EM_GETOLEINTERFACE, 0, (LPARAM)&lpREO ) ) {
                REOBJECT    stObj;

                stWTI.dwOLEObjects = lpREO->lpVtbl->GetObjectCount( lpREO );
                stWTI.dwOLELinks = lpREO->lpVtbl->GetLinkCount( lpREO );

                if( wStreamFlags & SFF_SELECTION ) {
                    /* Discard objects outside selected text */
                    DWORD   dwObjIdx, dwObjCnt, dwTmp;

                    dwObjCnt = stWTI.dwOLEObjects;
                    for( dwObjIdx=0; dwObjIdx<dwObjCnt; dwObjIdx++ ) {
                        dwTmp = (DWORD)lpREO->lpVtbl->GetObject( lpREO,
                            dwObjIdx,
                            &stObj,
                            REO_GETOBJ_NO_INTERFACES );
                        if( dwTmp == 0 ) {
                            dwTmp = stObj.cp;
                            if( (dwTmp < stCR.cpMin)||(dwTmp >= stCR.cpMax) )
                                stWTI.dwOLEObjects--;
                        }
                    }
                }

                lpREO->lpVtbl->Release( lpREO );
            }
            /* Stream text out of control to get more info */
            stWTI.dwLengthNoBlanks = stWTI.dwLength;
            stWTI.dwParagraphs = 0;
            stWTI.dwWords = 0;
            stWTI.boInWord = FALSE;
            stWTI.boInPara = FALSE;
            stWTI.hWndProgress = hWndProgress;
            stWTI.dwCurChar = 0;
            stWTI.dwProgressStepSize = stWTI.dwLength / PROGRESS_STEPS;
            stES.dwCookie = (DWORD)&stWTI;
            stES.pfnCallback = subStreamOutProc;
            SendMessage( hWndText, EM_STREAMOUT, wStreamFlags, (LPARAM)&stES );
            SendMessage( hWndProgress, PBM_SETPOS, PROGRESS_STEPS-1, 0 );
            /* Show results */
            subTextInfo( hWnd, IDT_LENGTH, stWTI.dwLength );
            subTextInfo( hWnd, IDT_CHARS, stWTI.dwLengthNoBlanks );
            subTextInfo( hWnd, IDT_PARAS, stWTI.dwParagraphs );
            subTextInfo( hWnd, IDT_WORDS, stWTI.dwWords );
            subTextInfo( hWnd, IDT_LINES, stWTI.dwLines );
            subTextInfo( hWnd, IDT_OLE, stWTI.dwOLEObjects );
            /* Hide progress bar and enable OK button */
            ShowWindow( hWndProgress, SW_HIDE );
            ShowWindow( hWndOkButton, SW_SHOW );
            SetFocus( hWndOkButton );
            /* Restore cursor and exit */
            SetCursor( hCursor );
            break;
        case WM_COMMAND:
            /* Not much to do with only one button */
            EndDialog( hWnd, 0 );
            break;
    }
    return( FALSE );
}

/*----------------------------------------------------------

    About information dialog box.

----------------------------------------------------------*/
BOOL CALLBACK AboutDlgProc( HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam )
{
    NOREF(wParam);
    NOREF(lParam);

    switch( wMsg ) {
        case WM_INITDIALOG:
            return( TRUE );
        case WM_COMMAND:
            EndDialog( hWnd, 0 );
            break;
    }
    return( FALSE );
}

/*----------------------------------------------------------

    Window procedure for the subclassed WordPad window.

----------------------------------------------------------*/
LRESULT CALLBACK wpxWndProc( HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam )
{
    switch( wMsg ) {
        /* Modify host GUI */
        case PM_INITGUI:
            /* Insert the new "tools" submenu into the menu bar*/
            hMenuMain = GetMenu( hWnd );
            hMenuTools = CreatePopupMenu();
            AppendMenu( hMenuTools, MF_STRING, IDM_TEXTINFO, "Text &Information" );
            AppendMenu( hMenuTools, MF_STRING, IDM_BACKCOLOR, "&Background Color" );
            AppendMenu( hMenuTools, MF_SEPARATOR, 0, NULL );
            AppendMenu( hMenuTools, MF_STRING, IDM_ABOUT, "&About WPX" );
            if( hMenuTools != NULL ) {
                InsertMenu( hMenuMain,
                    5,
                    MF_BYPOSITION | MF_POPUP,
                    (UINT)hMenuTools,
                    "&Tools" );
            }
            DrawMenuBar( hWnd );
            /* Search for text and status bar windows */
            {
                DWORD   dwMaxY = 0;

                EnumChildWindows( hWnd, subEnumChildProc, (LPARAM)&dwMaxY );
            }
            /* Get color information */
            rgbTextBackground = GetSysColor( COLOR_WINDOW );
            {
                int     iIdx;

                for( iIdx=0; iIdx<16; iIdx++ )
                    rgbCustomColors[iIdx] = rgbTextBackground;
            }
            return( 0 );
        /* Menu selection */
        case WM_MENUSELECT:
            if( (HMENU)lParam == hMenuTools ) {
                /* Our menu */
                switch( LOWORD(wParam) ) {
                    case IDM_TEXTINFO:
                        subStatusText( IDS_HELPTEXTINFO );
                        break;
                    case IDM_BACKCOLOR:
                        subStatusText( IDS_HELPBACKCOLOR );
                        break;
                    case IDM_ABOUT:
                        subStatusText( IDS_HELPABOUT );
                        break;
                }
                return( 0 );
            }
            if( (HMENU)lParam == hMenuMain ) {
                /* Main menu, clear status bar */
                subStatusText( 0 );
                return( 0 );
            }
            break;
        /* Popup menu initialization, hide our menu */
        /* from WordPad, or it will disable all items! */
        case WM_INITMENUPOPUP:
            if( (HMENU)wParam == hMenuTools )
                return( 0 );
            break;
        /* Menu command */
        case WM_COMMAND:
            switch( wParam ) {
                case IDM_TEXTINFO:
                    /* Display text information */
                    DialogBox( hInst,
                        MAKEINTRESOURCE(DLG_TEXTINFO),
                        hWnd,
                        TextInfoDlgProc );
                    return( 0 );
                case IDM_BACKCOLOR:
                    /* Changes the text background color */
                    {
                        CHOOSECOLOR stCC;

                        stCC.lStructSize = sizeof(stCC);
                        stCC.hwndOwner = hWndMain;
                        stCC.Flags = CC_RGBINIT;
                        stCC.rgbResult = rgbTextBackground;
                        stCC.lpCustColors = rgbCustomColors;
                        if( ChooseColor( &stCC ) ) {
                            rgbTextBackground = stCC.rgbResult;
                            SendMessage( hWndText,
                                EM_SETBKGNDCOLOR,
                                FALSE,
                                rgbTextBackground );
                        }
                    }
                    return( 0 );
                case IDM_ABOUT:
                    /* Display about box */
                    DialogBox( hInst,
                        MAKEINTRESOURCE(DLG_ABOUT),
                        hWnd,
                        AboutDlgProc );
                    return( 0 );
                /* Default: pass command to application */
            }
            break;
    }
    return( CallWindowProc( pfOldWndProc, hWnd, wMsg, wParam, lParam ) );
}

/*
    Returns TRUE if the specified window is of class
    WORDPADCLASSNAME, FALSE otherwise.
*/
BOOL IsWordPadWindow( HWND hWnd )
{
    CHAR    szBuf[128];

    if( (hAppWnd != NULL) )
        if( GetClassName( hWnd, szBuf, sizeof(szBuf) ) > 0 )
            if( lstrcmp( szBuf, WORDPADCLASSNAME ) == 0 )
                return( TRUE );
    return( FALSE );
}

/*
    Hook for the WM_SHELL class.

    Input:
        nCode   = hook code
        wParam  = additional info
        lParam  = additional info
    Output:
        zero

    Note: we use this function to detect when WORDPAD instances
    create and destroy main windows.
*/
LRESULT CALLBACK wpxShellHook( int nCode, WPARAM wParam, LPARAM lParam )
{
    switch( nCode ) {
        case HSHELL_WINDOWCREATED:
            if( IsWordPadWindow( (HWND)wParam ) ) {
                /* Subclass new window */
                hWndMain = (HWND)wParam;
                pfOldWndProc = (WNDPROC)SetWindowLong( hWndMain,
                    GWL_WNDPROC,
                    (LONG)wpxWndProc );
                /* Notify application */
                PostMessage( hAppWnd, PM_INSERTWINDOW, wParam, TRUE );
            }
            break;
        case HSHELL_WINDOWDESTROYED:
            if( IsWordPadWindow( (HWND)wParam ) ) {
                /* Remove subclassing */
                SetWindowLong( (HWND)wParam,
                    GWL_WNDPROC,
                    (LONG)pfOldWndProc );
                /* Notify application */
                PostMessage( hAppWnd, PM_REMOVEWINDOW, wParam, 0 );
            }
            break;
    }

    return( CallNextHookEx( hShlHook, nCode, wParam, lParam ) );
}

/*
    Hook for the WM_GETMESSAGE class.

    Input:
        nCode   = hook code
        wParam  = removal flag
        lParam  = pointer to message structure
    Output:
        zero

    Note: we use this hook to lock the DLL inside the WORDPAD
    address space, and to start subclassing.
*/
LRESULT CALLBACK wpxMsgHook( int nCode, WPARAM wParam, LPARAM lParam )
{
    if( nCode == HC_ACTION ) {
        if( wParam == PM_REMOVE ) {
            MSG *   pMsg;

            pMsg = (MSG *)lParam;
            if( pMsg->message == PM_SUBCLASSWINDOW ) {
                pfOldWndProc = (WNDPROC)SetWindowLong( pMsg->hwnd,
                    GWL_WNDPROC,
                    (LONG)wpxWndProc );
                PostMessage( hAppWnd, PM_INSERTWINDOW, (WPARAM)pMsg->hwnd, FALSE );
                return( TRUE );
            }
        }
        return( FALSE );
    }

    return( CallNextHookEx( hMsgHook, nCode, wParam, lParam ) );
}

/*
    Installs the WM_SHELL hook function.

    Input:
        none
    Output:
        FALSE on success, TRUE otherwise
*/
HHOOK CALLBACK wpxInitShellHook( DWORD dwThreadId )
{
    if( hShlHook != NULL )
        return( FALSE );

    hAppWnd = FindWindow( WPXCLASSNAME, WPXWINDOWNAME );
    if( hAppWnd == NULL )
        return( NULL );
    hShlHook = SetWindowsHookEx( WH_SHELL,
        wpxShellHook,
        hInst,
        dwThreadId );
    return( hShlHook );
}

/*
    Uninstalls the WM_SHELL hook function.

    Input:
        none
    Output:
        FALSE on success, TRUE otherwise
*/
BOOL CALLBACK wpxTermShellHook( void )
{
    if( hShlHook != NULL ) {
        UnhookWindowsHookEx( hShlHook );
        hShlHook = NULL;
    }

    return( FALSE );
}

/*
    Installs the WM_GETMESSAGE hook function.

    Input:
        none
    Output:
        FALSE on success, TRUE otherwise
*/
HHOOK CALLBACK wpxInitMsgHook( DWORD dwThreadId )
{
    if( hMsgHook != NULL )
        return( FALSE );

    hMsgHook = SetWindowsHookEx( WH_GETMESSAGE,
        wpxMsgHook,
        hInst,
        dwThreadId );

    return( hMsgHook );
}

/*
    Uninstalls the WM_GETMESSAGE hook function.

    Input:
        none
    Output:
        FALSE on success, TRUE otherwise
*/
BOOL CALLBACK wpxTermMsgHook( void )
{
    if( hMsgHook != NULL ) {
        UnhookWindowsHookEx( hMsgHook );
        hMsgHook = NULL;
    }

    return( FALSE );
}

/*
    DLL entry point.

    Input:
        hInstDLL    = DLL module handle
        dwReason    = reason why the function is called
        lpvRes      = pointer to extra information (reserved)
    Output:
        always TRUE
*/
BOOL APIENTRY DllEntryPoint( HINSTANCE hInstDLL, DWORD dwReason, LPVOID lpvRes )
{
    NOREF(lpvRes);

    switch( dwReason ) {
        case DLL_PROCESS_ATTACH:
            /* Whenever a new instance of this DLL is created, */
            /* we must find the control window and store */
            /* the handle in this instance data */
            hInst = hInstDLL;
            hAppWnd = FindWindow( WPXCLASSNAME, WPXWINDOWNAME );
            if( hAppWnd != NULL ) {
                /* Not the first instance, get the hook handle */
                hMsgHook = (HHOOK)SendMessage( hAppWnd, PM_GETMSGHOOK, 0, 0 );
                hShlHook = (HHOOK)SendMessage( hAppWnd, PM_GETSHELLHOOK, 0, 0 );
            }
            break;
        case DLL_PROCESS_DETACH:
            break;
    }

    return( TRUE );
}
